/**
 * Copyright Copyright 2008-14 Simon Andrews and Curtis Rueden
 *
 *    This file is part of FocalPoint.
 *
 *    FocalPoint is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    FocalPoint is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with StackMeasure; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package uk.ac.bbsrc.babraham.FocalPoint.GUI.Thumbnail;

import ij.ImagePlus;
import ij.ImageStack;
import ij.process.ImageProcessor;
import ij.CompositeImage;

import java.io.File;
import java.io.IOException;

import loci.plugins.util.ImageProcessorReader;


public class BioFormatsThumbnailPanel extends ThumbnailPanel  {

	
	/*
	 * I'd prefer not to try to assign file types by extension, but unfortunately
	 * the isThisType method from BioFormats is *really* slow on big files it can't
	 * read (30 seconds plus to say it can't open a 1G tar.gz file, so I'm going to
	 * have to filter by extension instead.
	 * 
	 */

	private static final String [] extensions = new String [] {
		".psd",
		".al3d",
		".gel",
		".tif",
		".tiff",
		".svs",
		".avi",
		".sdt",
		".pic",
		".ims",
		".bmp",
		".dv",
		".r3d",
		".dcm",
		".dicom",
		".eps",
		".flex",
		".flx", // guessing here...
		".fits",
		".d3m",
		".gif",
		".ics",
		".seq",
		".ipw",
		".liff",
		".lif", //guessing here...
		".lei",
		".l2d",
		".scn",
		".lim",
		".stk",
//		".txt",
		".mng",
		".mrc",
		".nef",
		".nd2",
		".nrrd",
		".sld",
		".oib",
		".oif",
		".ome",
		".pict",
		".pgm",
		".png",
		".mov",
		".cxd",
		".xys",
		".zvi",
		".lsm",
		".fts",
		".czi",
	};
	
	public static boolean imageLoading = false;

	private ImageProcessorReader reader;
	
	public BioFormatsThumbnailPanel (FileDetailsPanel fileDetails) throws IncompatibleFileTypeException {
		super(fileDetails);
	}


	public ImagePlus [] loadImages (File file) throws IOException {
						
		try {
			if (reader == null) makeReader();
			reader.setId(file.getAbsolutePath());
			
			
			/*
			 * BioFormats image channels are a mix of separate
			 * collection channels and colour channels.
			 * 
			 * getSizeC gives you the total number of channels
			 * and getRBGChannel count tells you how many colour
			 * channels there are.  EffectiveSizeC is the total
			 * channels divided by the colour channels.
			 * 
			 */
			
			int seriesCount = reader.getEffectiveSizeC();
			
			System.err.println("There are "+seriesCount+" series");

			ImagePlus[] ip = new ImagePlus[seriesCount];
			ImageStack[] stacks = new ImageStack[seriesCount];
			
			channels = ip.length;
			
			for (int s=0;s<stacks.length;s++) {
				stacks[s] = new ImageStack(reader.getSizeX(),reader.getSizeY());
			}
			
			for (int i=0;i<reader.getImageCount();i++) {
				// read image plane

				ImageProcessor [] procs = reader.openProcessors(i);
				
				ImageStack stack = new ImageStack(reader.getSizeX(), reader.getSizeY());
				for (ImageProcessor p : procs) {
					stack.addSlice("", p);
				}
				ImagePlus rawImp = new ImagePlus("", stack);
				ImagePlus colorizedImp = null;
				 
				if (procs.length > 1) {
					colorizedImp = new CompositeImage(rawImp, CompositeImage.COMPOSITE);
				}
				else colorizedImp = rawImp;
				
				// determine channel for this plane and add to appropriate stack
				int[] zct = reader.getZCTCoords(i);
				int c = zct[1]/reader.getRGBChannelCount();

//				System.err.println("From file "+file.getName()+" and image "+i+" I got stack "+c+" with "+procs.length+" procs");
				
				
				/*
				 * TODO: There is a problem here when using CMYK images (eg demo confocal.ics)
				 * ImageJ will correctly display this in colour, but my JImagePanel shows it
				 * as black and white.  I may need to modify a newer version of ImagePanel.
				 */

				stacks[c].addSlice("", colorizedImp.getProcessor());
			}
			reader.close();
			// Temporary - trying to find memory leaks!
			reader = null;

			for (int s=0;s<stacks.length;s++) {
				ip[s] = new ImagePlus(file.getName(),stacks[s]);
			}
			imageLoading = false;
			return ip;
		} 
		catch (Exception e) {
			e.printStackTrace();
			imageLoading = false;
			throw new IOException(e.getLocalizedMessage());
		}
	}


	long getFileSize(File file) {
		if (file.getName().toLowerCase().endsWith("oif")) {
			/*
			 * A bit more complex.  The actual file size is the
			 * cumulative size of the .roi files in the directory
			 * associated with this file.
			 */

			File dir = new File(file.getAbsoluteFile()+".files");
			if (!dir.exists()) {
				System.out.println("Can't find the directory associated with "+file.getName());
				return 0;
			}
			File [] subfiles = dir.listFiles();
			long totalSize = file.length();
			for (int i=0;i<subfiles.length;i++) {
				totalSize += subfiles[i].length();
			}
			return totalSize;
		}
		
		//TODO: Bioformats has a routine already to specify the size of all the
		// files used when loading an image.  This more generic approach would
		// be more appropriate to use here, but may not be possible until we've
		// actually created the format reader to load the file.  We'd therefore
		// need to be able to update the file sizes.

		// Nice and straightforward, just give the normal file size
		return file.length();
	}


	boolean isImage() {
		return true;
	}


	boolean shouldDeferLoading(File file) {
		// 20MB total size limit for this file.
		if (getFileSize(file) > (20*1024*1024)) {
			return true;
		}
		return false;
	}


	boolean canReadFile(File file) {
		if (reader == null) makeReader();
		String name = file.getName().toLowerCase();
		for (int i=0;i<extensions.length;i++) {
			if (name.endsWith(extensions[i])) {
				// We might support it
				return reader.isThisType(fileDetails.file().getAbsolutePath());				
			}
		}
		return false;
	}

	private void makeReader() {
		reader = new ImageProcessorReader();
	}
		
}
